@@ -39,6 +39,7 @@
 #include <dirent.h>
 #include <string.h>
 #include <stdio.h>
+#include <time.h>
 #include "synproto.h"
 #include "synapticsstr.h"
 #include <xf86.h>
@@ -80,6 +81,8 @@ struct eventcomm_proto_data {
     ValuatorMask **last_mt_vals;
     int num_touches;
     int *tracking_ids;
+
+    int have_monotonic_clock;
 };
 
 struct eventcomm_proto_data *
@@ -198,11 +201,11 @@ EventDeviceOnHook(InputInfoPtr pInfo, SynapticsParameters * para)
     SynapticsPrivate *priv = (SynapticsPrivate *) pInfo->private;
     struct eventcomm_proto_data *proto_data =
         (struct eventcomm_proto_data *) priv->proto_data;
+    int clockid = CLOCK_MONOTONIC;
+    int ret;
 
     if (para->grab_event_device) {
         /* Try to grab the event device so that data don't leak to /dev/input/mice */
-        int ret;
-
         SYSCALL(ret = ioctl(pInfo->fd, EVIOCGRAB, (pointer) 1));
         if (ret < 0) {
             xf86IDrvMsg(pInfo, X_WARNING, "can't grab event device, errno=%d\n",
@@ -213,6 +216,11 @@ EventDeviceOnHook(InputInfoPtr pInfo, SynapticsParameters * para)
 
     proto_data->need_grab = FALSE;
 
+#ifdef EVIOCSCLOCKID
+    SYSCALL(ret = ioctl(pInfo->fd, EVIOCSCLOCKID, &clockid));
+    proto_data->have_monotonic_clock = (ret == 0);
+#endif
+
     InitializeTouch(pInfo);
 
     return TRUE;
@@ -684,7 +692,10 @@ EventReadHwState(InputInfoPtr pInfo,
             switch (ev.code) {
             case SYN_REPORT:
                 hw->numFingers = count_fingers(pInfo, comm);
-                hw->millis = 1000 * ev.time.tv_sec + ev.time.tv_usec / 1000;
+                if (proto_data->have_monotonic_clock)
+                    hw->millis = 1000 * ev.time.tv_sec + ev.time.tv_usec / 1000;
+                else
+                    hw->millis = GetTimeInMillis();
                 SynapticsCopyHwState(hwRet, hw);
                 return TRUE;
             }
@@ -854,6 +865,9 @@ event_query_touch(InputInfoPtr pInfo)
             AXIS_LABEL_PROP_ABS_MT_BLOB_ID,
             AXIS_LABEL_PROP_ABS_MT_TRACKING_ID,
             AXIS_LABEL_PROP_ABS_MT_PRESSURE,
+            AXIS_LABEL_PROP_ABS_MT_DISTANCE,
+            AXIS_LABEL_PROP_ABS_MT_TOOL_X,
+            AXIS_LABEL_PROP_ABS_MT_TOOL_Y,
         };
 
         if (mtdev->caps.slot.maximum > 0)
@@ -884,8 +898,14 @@ event_query_touch(InputInfoPtr pInfo)
                 case ABS_MT_TRACKING_ID - ABS_MT_TOUCH_MAJOR:
                     break;
 
-                default:
-                    priv->touch_axes[axnum].label = labels[i];
+                    if (i >= sizeof(labels)/sizeof(labels[0])) {
+                        xf86IDrvMsg(pInfo, X_ERROR,
+                                    "Axis %d out of label range. This is a bug\n",
+                                    i);
+                        priv->touch_axes[axnum].label = NULL;
+                    } else
+                        priv->touch_axes[axnum].label = labels[i];
+
                     priv->touch_axes[axnum].min = mtdev->caps.abs[i].minimum;
                     priv->touch_axes[axnum].max = mtdev->caps.abs[i].maximum;
                     /* Kernel provides units/mm, X wants units/m */
